@@ -1,9 +1,16 @@
 
+#include "include/stringify.h"
 #include "CrushTester.h"
+#include "CrushTreeDumper.h"
 
 #include <algorithm>
 #include <stdlib.h>
-
+/* fork */
+#include <unistd.h>
+/* waitpid */
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <common/errno.h>
 
 void CrushTester::set_device_weight(int dev, float f)
 {
@@ -350,6 +357,139 @@ void CrushTester::write_integer_indexed_scalar_data_string(vector<string> &dst,
   dst.push_back( data_buffer.str() );
 }
 
+int CrushTester::test_with_crushtool(const string& crushtool,
+                                     int max_id,
+                                     int timeout)
+{
+  string timeout_string = stringify(timeout);
+  string opt_max_id = stringify(max_id);
+  vector<const char *> cmd_args;
+  cmd_args.push_back("timeout");
+  cmd_args.push_back(timeout_string.c_str());
+  cmd_args.push_back(crushtool.c_str());
+  cmd_args.push_back("-i");
+  cmd_args.push_back("-");
+  cmd_args.push_back("--test");
+  cmd_args.push_back("--check");
+  cmd_args.push_back(opt_max_id.c_str());
+  cmd_args.push_back(NULL);
+
+  int pipefds[2];
+  if (::pipe(pipefds) == -1) {
+    int r = errno;
+    err << "error creating pipe: " << cpp_strerror(r) << "\n";
+    return -r;
+  }
+
+  int fpid = fork();
+  if (fpid < 0) {
+    int r = errno;
+    err << "unable to fork(): " << cpp_strerror(r);
+    ::close(pipefds[0]);
+    ::close(pipefds[1]);
+    return -r;
+  } else if (fpid == 0) {
+    ::close(pipefds[1]);
+    ::dup2(pipefds[0], STDIN_FILENO);
+    ::close(pipefds[0]);
+    ::close(1);
+    ::close(2);
+    int r = execvp(cmd_args[0], (char * const *)&cmd_args[0]);
+    if (r < 0)
+      exit(errno);
+    // we should never reach this
+    exit(EINVAL);
+  }
+  ::close(pipefds[0]);
+
+  bufferlist bl;
+  ::encode(crush, bl);
+  bl.write_fd(pipefds[1]);
+  ::close(pipefds[1]);
+
+  int status;
+  int r = waitpid(fpid, &status, 0);
+  assert(r == fpid);
+
+  if (!WIFEXITED(status)) {
+    assert(WIFSIGNALED(status));
+    err << "error testing crush map\n";
+    return -EINVAL;
+  }
+
+  r = WEXITSTATUS(status);
+  if (r == 0) {
+    // major success!
+    return 0;
+  }
+  if (r == 124) {
+    // the test takes longer than timeout and was interrupted
+    return -EINTR;
+  }
+
+  if (r == ENOENT) {
+    err << "unable to find " << cmd_args << " to test the map";
+    return -ENOENT;
+  }
+
+  // something else entirely happened
+  // log it and consider an invalid crush map
+  err << "error running crushmap through crushtool: " << cpp_strerror(r);
+  return -EINVAL;
+}
+
+namespace {
+  class BadCrushMap : public std::runtime_error {
+  public:
+    int item;
+    BadCrushMap(const char* msg, int id)
+      : std::runtime_error(msg), item(id) {}
+  };
+  // throws if any node in the crush fail to print
+  class CrushWalker : public CrushTreeDumper::Dumper<void> {
+    typedef void DumbFormatter;
+    typedef CrushTreeDumper::Dumper<DumbFormatter> Parent;
+    unsigned max_id;
+  public:
+    CrushWalker(const CrushWrapper *crush, unsigned max_id)
+      : Parent(crush), max_id(max_id) {}
+    void dump_item(const CrushTreeDumper::Item &qi, DumbFormatter *) {
+      int type = -1;
+      if (qi.is_bucket()) {
+	if (!crush->get_item_name(qi.id)) {
+	  throw BadCrushMap("unknown item name", qi.id);
+	}
+	type = crush->get_bucket_type(qi.id);
+      } else {
+	if (max_id > 0 && qi.id >= (int)max_id) {
+	  throw BadCrushMap("item id too large", qi.id);
+	}
+	type = 0;
+      }
+      if (!crush->get_type_name(type)) {
+	throw BadCrushMap("unknown type name", qi.id);
+      }
+    }
+  };
+}
+
+bool CrushTester::check_name_maps(unsigned max_id) const
+{
+  CrushWalker crush_walker(&crush, max_id);
+  try {
+    // walk through the crush, to see if its self-contained
+    crush_walker.dump(NULL);
+    // and see if the maps is also able to handle straying OSDs, whose id >= 0.
+    // "ceph osd tree" will try to print them, even they are not listed in the
+    // crush map.
+    crush_walker.dump_item(CrushTreeDumper::Item(0, 0, 0), NULL);
+  } catch (const BadCrushMap& e) {
+    err << e.what() << ": item#" << e.item << std::endl;
+    return false;
+  }
+  return true;
+}
+
 int CrushTester::test()
 {
   if (min_rule < 0 || max_rule < 0) {
