@@ -367,26 +367,6 @@ public:
   const pg_pool_t &get_pool() const {
     return pool.info;
   }
-  map<pg_shard_t, const pg_missing_t *> get_all_missing() const {
-    map<pg_shard_t, const pg_missing_t *> all_missing;
-    all_missing[pg_whoami] = &(pg_log.get_missing());
-    for (map<pg_shard_t, pg_missing_t>::const_iterator i = peer_missing.begin();
-	 i != peer_missing.end();
-	 ++i) {
-      all_missing[i->first] = &(i->second);
-    }
-    return all_missing;
-  }
-  const map<pg_shard_t, const pg_info_t *> get_all_info() const {
-    map<pg_shard_t, const pg_info_t *> all_info;
-    all_info[pg_whoami] = &info;
-    for (map<pg_shard_t, pg_info_t>::const_iterator i = peer_info.begin();
-	 i != peer_info.end();
-	 ++i) {
-      all_info[i->first] = &(i->second);
-    }
-    return all_info;
-  }
   ObjectContextRef get_obc(
     const hobject_t &hoid,
     map<string, bufferlist> &attrs) {
@@ -415,13 +395,11 @@ public:
     if (peer == get_primary())
       return true;
     assert(peer_info.count(peer));
-    bool should_send_backfill = hoid.pool != (int64_t)info.pgid.pool() ||
+    bool should_send = hoid.pool != (int64_t)info.pgid.pool() ||
       hoid <= MAX(last_backfill_started, peer_info[peer].last_backfill);
-    if (!should_send_backfill)
+    if (!should_send)
       assert(is_backfill_targets(peer));
-
-    assert(peer_missing.count(peer));
-    return should_send_backfill && !peer_missing[peer].is_missing(hoid);
+    return should_send;
   }
   
   void update_peer_last_complete_ondisk(
@@ -874,7 +852,20 @@ protected:
 	requeue_snaptrimmer_clone ||
 	requeue_snaptrimmer_snapset)
       queue_snap_trim();
-    requeue_ops(to_req);
+
+    if (!to_req.empty()) {
+      assert(ctx->obc);
+      // requeue at front of scrub blocking queue if we are blocked by scrub
+      if (scrubber.write_blocked_by_scrub(ctx->obc->obs.oi.soid.get_head())) {
+	waiting_for_active.splice(
+	  waiting_for_active.begin(),
+	  to_req,
+	  to_req.begin(),
+	  to_req.end());
+      } else {
+	requeue_ops(to_req);
+      }
+    }
   }
 
   // replica ops
@@ -887,7 +878,7 @@ protected:
   void repop_all_applied(RepGather *repop);
   void repop_all_committed(RepGather *repop);
   void eval_repop(RepGather*);
-  void issue_repop(RepGather *repop, utime_t now);
+  void issue_repop(RepGather *repop);
   RepGather *new_repop(OpContext *ctx, ObjectContextRef obc, ceph_tid_t rep_tid);
   void remove_repop(RepGather *repop);
 
@@ -939,7 +930,8 @@ protected:
   /// clear agent state
   void agent_clear();
 
-  void agent_choose_mode(bool restart = false);  ///< choose (new) agent mode(s)
+  /// choose (new) agent mode(s), returns true if op is requeued
+  bool agent_choose_mode(bool restart = false, OpRequestRef op = OpRequestRef());
   void agent_choose_mode_restart();
 
   /// true if we can send an ondisk/commit for v
@@ -973,8 +965,6 @@ protected:
     return true;
   }
 
-  friend struct C_OnPushCommit;
-
   // projected object info
   SharedLRU<hobject_t, ObjectContext> object_contexts;
   // map from oid.snapdir() to SnapSetContext *
@@ -1169,7 +1159,10 @@ protected:
    */
   void do_cache_redirect(OpRequestRef op);
   /**
-   * This function starts up a copy from
+   * This function attempts to start a promote.  Either it succeeds,
+   * or places op on a wait list.  If op is null, failure means that
+   * this is a noop.  If a future user wants to be able to distinguish
+   * these cases, a return value should be added.
    */
   void promote_object(ObjectContextRef obc,            ///< [optional] obc
 		      const hobject_t& missing_object, ///< oid (if !obc)
@@ -1499,6 +1492,9 @@ private:
 
   int _verify_no_head_clones(const hobject_t& soid,
 			     const SnapSet& ss);
+  // return true if we're creating a local object, false for a
+  // whiteout or no change.
+  bool maybe_create_new_object(OpContext *ctx);
   int _delete_oid(OpContext *ctx, bool no_whiteout);
   int _rollback_to(OpContext *ctx, ceph_osd_op& op);
 public:
@@ -1511,11 +1507,6 @@ public:
   void wait_for_all_missing(OpRequestRef op);
 
   bool is_degraded_or_backfilling_object(const hobject_t& oid);
-
-  /* true if the object is missing on any peer, *healthy_copies will be
-   * set to the number of complete peers not missing the object
-   */
-  bool is_degraded_object(const hobject_t &oid, int *healthy_copies);
   void wait_for_degraded_object(const hobject_t& oid, OpRequestRef op);
 
   bool maybe_await_blocked_snapset(const hobject_t &soid, OpRequestRef op);
